"""
This module implements the IPADE_ID method.
"""
from dataclasses import dataclass

import numpy as np

from sklearn.metrics import roc_auc_score, accuracy_score
from sklearn.preprocessing import MinMaxScaler

from ..base import OverSampling
from ..base import mode, instantiate_obj

from .._logger import logger
_logger= logger

__all__= ['IPADE_ID']

@dataclass
class TrainingSet:
    """
    Represents a trining set
    """
    X: np.array
    y: np.array

class IPADE_ID(OverSampling):
    """
    References:

        * BibTex::

            @article{ipade_id,
                    title = "Addressing imbalanced classification with
                                instance generation techniques: IPADE-ID",
                    journal = "Neurocomputing",
                    volume = "126",
                    pages = "15 - 28",
                    year = "2014",
                    note = "Recent trends in Intelligent Data Analysis Online
                                Data Processing",
                    issn = "0925-2312",
                    doi = "https://doi.org/10.1016/j.neucom.2013.01.050",
                    author = "Victoria López and Isaac Triguero and Cristóbal
                                J. Carmona and Salvador García and
                                Francisco Herrera",
                    keywords = "Differential evolution, Instance generation,
                                Nearest neighbor, Decision tree, Imbalanced
                                datasets"
                    }

    Notes:

        * According to the algorithm, if the addition of a majority sample
            doesn't improve the AUC during the DE optimization process,
            the addition of no further majority points is tried.
        * In the differential evolution the multiplication by a random number
            seems have a deteriorating effect, new scaling parameter added to
            fix this.
        * It is not specified how to do the evaluation.
        * When the dt classifier is grown to full depth, each leaf contains
            only one element, and there are no independent elements left
            for validation.
    """

    categories = [OverSampling.cat_changes_majority,
                  OverSampling.cat_memetic,
                  OverSampling.cat_uses_classifier]

    def __init__(self,
                 *,
                 F=0.1,
                 G=0.1,
                 OT=20,
                 max_it=40,
                 dt_classifier=('sklearn.tree',
                                'DecisionTreeClassifier',
                                {'random_state': 2}),
                 base_classifier=('sklearn.tree',
                                'DecisionTreeClassifier',
                                {'random_state': 2}),
                 n_jobs=1,
                 random_state=None,
                 **_kwargs):
        """
        Constructor of the sampling object

        Args:
            F (float): control parameter of differential evolution
            G (float): control parameter of the evolution
            OT (int): number of optimizations
            max_it (int): maximum number of iterations for DE_optimization
            dt_classifier (obj): decision tree classifier object
            base_classifier (obj): classifier object
            n_jobs (int): number of parallel jobs
            random_state (int/RandomState/None): initializer of random_state,
                                                    like in sklearn
        """
        super().__init__(random_state=random_state, checks={'min_n_min': 4})
        self.check_greater(F, 'F', 0)
        self.check_greater(G, 'G', 0)
        self.check_greater(OT, 'OT', 0)
        self.check_greater(max_it, 'max_it', 0)
        self.check_n_jobs(n_jobs, 'n_jobs')

        self.de_params = {'F': F,
                            'G': G,
                            'OT': OT,
                            'max_it': max_it,
                            'dt_classifier': dt_classifier,
                            'base_classifier': base_classifier}

        self.n_jobs = n_jobs

        self.dt_classifier_obj = instantiate_obj(dt_classifier)
        self.base_classifier_obj = instantiate_obj(base_classifier)

    @ classmethod
    def parameter_combinations(cls, raw=False):
        """
        Generates reasonable parameter combinations.

        Returns:
            list(dict): a list of meaningful parameter combinations
        """
        # as the OT and max_it parameters control the discovery of the feature
        # space it is enough to try sufficiently large numbers
        dt_classifiers = [('sklearn.tree',
                            'DecisionTreeClassifier',
                            {'random_state': 2})]
        base_classifiers = [('sklearn.tree',
                            'DecisionTreeClassifier',
                            {'random_state': 2})]
        parameter_combinations = {'F': [0.1, 0.2],
                                  'G': [0.1, 0.2],
                                  'OT': [30],
                                  'max_it': [40],
                                  'dt_classifier': dt_classifiers,
                                  'base_classifier': base_classifiers}
        return cls.generate_parameter_combinations(parameter_combinations, raw)

    def update_element(self, *, GS_y, idx, GS_x, indices, TR_X):
        """
        Updates the GS_idx vector

        Args:
            GS_y (np.array): the y labels in the optimization set
            idx (np.array): index of the vector to update
            GS_x (np.array): the vector to update
            indices (dict): indices of the various classes
            TR_X (np.array): the original vectors

        Returns:
            np.array: the updated GS_x vector
        """
        rand_idx = self.random_state.choice(indices[GS_y[idx]],
                                            3,
                                            replace=False)

        random = self.random_state.random_sample()
        value = GS_x + self.de_params['G'] * random\
                                        * (TR_X[rand_idx[0]] - GS_x)\
                            + self.de_params['F']\
                                        * (TR_X[rand_idx[1]] - TR_X[rand_idx[2]])

        return value

    def DE_optimization(self, # pylint: disable=invalid-name
                            *,
                            GS, # pylint: disable=invalid-name
                            TR,
                            indices,
                            for_validation):
        """
        Implements the DE_optimization method of the paper.
        Roughly, for each point of GS takes 3 random samples from the same
        class and moves the actual point towards them, then
        evaluates the new configuration.

        Args:
            GS (TrainingSet): the optimized training set
            TR (TrainingSet): the original training set
            indices (dict): indices of the various class labels
            for_validation (np.array): array of indices for X used for
                                        validation

        Returns:
            TrainingSet: optimized training set
        """

        # evaluate training set
        AUC_GS = self.evaluate_ID(GS=GS, # pylint: disable=invalid-name
                                    TR=TrainingSet(TR.X[for_validation],
                                                    TR.y[for_validation]))

        # optimizing the training set
        for _ in range(self.de_params['max_it']):
            GS_hat = [] # pylint: disable=invalid-name
            # doing the differential evolution
            for idx, GS_x in enumerate(GS.X): # pylint: disable=invalid-name
                value = self.update_element(GS_y=GS.y,
                                            idx=idx,
                                            GS_x=GS_x,
                                            indices=indices,
                                            TR_X=TR.X)

                GS_hat.append(np.clip(value, 0.0, 1.0))

            GS_hat = np.vstack(GS_hat) # pylint: disable=invalid-name

            # evaluating the current setting
            AUC_GS_hat = self.evaluate_ID(GS=TrainingSet(GS_hat, GS.y), # pylint: disable=invalid-name
                                        TR=TrainingSet(TR.X[for_validation],
                                                        TR.y[for_validation]))

            if AUC_GS_hat > AUC_GS:
                GS = TrainingSet(GS_hat, GS.y)
                AUC_GS = AUC_GS_hat # pylint: disable=invalid-name

        return GS

    def evaluate_ID(self, *, GS, TR): # pylint: disable=invalid-name
        """
        Implements the evaluate_ID function of the paper.

        Args:
            GS (TrainingSet): the optimized training set
            TR (TrainingSet): the original training set

        Returns:
            float: ROC AUC score
        """

        base_classifier = self.new_base_classifier()

        base_classifier.fit(GS.X, GS.y)
        min_index = np.where(base_classifier.classes_ == self.min_label)[0][0]
        pred = base_classifier.predict_proba(TR.X)[:, min_index]

        return roc_auc_score(TR.y, pred)

    def evaluate_class(self, *, GS, TR):
        """
        Implements the classification evaluation function of the paper.

        Args:
            GS (TrainingSet): the optimized training set
            TR (TrainingSet): the original training set

        Returns:
            float: accuracy score
        """
        base_classifier = self.new_base_classifier()

        base_classifier.fit(GS.X, GS.y)
        pred = base_classifier.predict(TR.X)

        return accuracy_score(TR.y, pred)

    def new_base_classifier(self):
        """
        Instantiates a new instance of the base classifier

        Returns:
            obj: a new instance of the base classifier
        """
        class_ = self.base_classifier_obj.__class__
        return class_(**(self.base_classifier_obj.get_params()))

    def configuration_checks(self,
                                y,
                                for_validation,
                                GS_y # pylint: disable=invalid-name
                                ):
        """
        Checks if the configuration is valid.

        Args:
            y (np.array): target labels
            for_validation (np.array): the indices of validation samples
            GS_y (np.array): the updated target labels
        """
        if len(np.unique(y[for_validation])) == 1:
            _logger.info("%s: No minority samples in validation set",
                        self.__class__.__name__)
            raise ValueError("No minority samples in validation set")

        if len(np.unique(GS_y)) == 1:
            _logger.info("%s: No minority samples in reduced dataset",
                            self.__class__.__name__)
            raise ValueError("No minority samples in reduced dataset")

    def phase_1(self,
                TR, # pylint: disable=invalid-name
                indices):
        """
        Initialization of the method.

        Args:
            TR (TrainingSet): the original training set
            indices (dict): the indices of the various classes

        Returns:
            float, np.array, TrainingSet: the AUC score, the validation
                                        indices and the initial optimized
                                        training set
        """
        _logger.info("%s: Initialization", self.__class__.__name__)

        self.dt_classifier_obj.fit(TR.X, TR.y)
        leafs = self.dt_classifier_obj.apply(TR.X)
        unique_leafs = np.unique(leafs)

        used_in_GS = np.repeat(False, len(TR.X)) # pylint: disable=invalid-name
        for_validation = np.where(np.logical_not(used_in_GS))[0]

        # extracting mean elements of the leafs
        GS_X = [] # pylint: disable=invalid-name
        GS_y = [] # pylint: disable=invalid-name
        for leaf in unique_leafs:
            leaf_indices = np.where(leafs == leaf)[0]
            GS_X.append(np.mean(TR.X[leaf_indices], axis=0))
            GS_y.append(mode(TR.y[leaf_indices]))
            if len(leaf_indices) == 1:
                used_in_GS[leaf_indices[0]] = True

        GS = TrainingSet(np.vstack(GS_X), np.array(GS_y)) # pylint: disable=invalid-name

        # updating the indices of the validation set excluding those used in GS
        for_validation = np.where(np.logical_not(used_in_GS))[0]

        _logger.info("%s: Size of validation set %d",
                    self.__class__.__name__, len(for_validation))
        self.configuration_checks(TR.y, for_validation, GS.y)


        # DE optimization takes place
        _logger.info("%s: DE optimization", self.__class__.__name__)

        GS = self.DE_optimization(GS=GS, # pylint: disable=invalid-name
                                TR=TR,
                                indices=indices,
                                for_validation=for_validation)

        # evaluate results
        AUC = self.evaluate_ID(GS=GS, # pylint: disable=invalid-name
                                TR=TrainingSet(TR.X[for_validation], TR.y[for_validation]))

        return AUC, for_validation, GS

    def determine_target_class(self, *, TR, GS,
                                for_validation, register_class):
        """
        Determine the target class according to the description in the paper

        Args:
            TR (TrainingSet): the original training set
            GS (TrainingSet): the optimized training set
            for_validation (np.array): labels used for validation
            register_class (dict): the status of the class

        Returns:
            int: the target class label
        """
        accuracy_class = {self.min_label: 0, self.maj_label: 0}

        less_accuracy = np.inf

        # loop in line 8
        for idx in [self.min_label, self.maj_label]:
            # condition in line 9
            if register_class[idx] == 'optimizable':
                y_mask = TR.y[for_validation] == idx
                class_for_validation = for_validation[y_mask]
                accuracy_class[idx] = self.evaluate_class(GS=GS,
                                                    TR=TrainingSet(TR.X[class_for_validation],
                                                                    TR.y[class_for_validation]))
                if accuracy_class[idx] < less_accuracy:
                    less_accuracy = accuracy_class[idx]
                    target_class = idx

        return target_class

    def generate_trial(self, *, indices, target_class, TR, GS, for_validation):
        """
        Generates a trial training set

        Args:
            indices (dict): the indices of the various classes
            target_class (int): the target class label
            TR (TrainingSet): the original training set
            GS (TrainingSet): the optimized training set
            for_validation (np.array): the indices used for validation

        Returns:
            TrainingSet, np.array: the trial training set and the updated
                                    validation indices
        """
        idx = self.random_state.choice(indices[target_class])

        GS_trial_X = np.vstack([GS.X, TR.X[idx]]) # pylint: disable=invalid-name
        GS_trial_y = np.hstack([GS.y, TR.y[idx]]) # pylint: disable=invalid-name

        # removing idx from the validation set in order to keep
        # the validation fair
        for_validation_trial = for_validation[for_validation != idx]

        return TrainingSet(GS_trial_X, GS_trial_y), for_validation_trial

    def phase_2(self, *, TR, indices, for_validation, GS, AUC):
        """
        Phase 2 of the algorithm

        Args:
            TR (TrainingSet): original training set
            indices (dict): indices of the various classes
            for_validation (np.array): the mask of the validation samples
            GS (TrainingSet): the optimized training set
            AUC (float): the AUC score

        Returns:
            TrainingSet: the optimized training set
        """
        register_class = {self.min_label: 'optimizable',
                          self.maj_label: 'optimizable'}
        number_of_optimizations = {self.min_label: 0,
                                   self.maj_label: 0}

        _logger.info("%s: Starting optimization", self.__class__.__name__)

        GS_trial = None # pylint: disable=invalid-name

        while (AUC < 1.0
                and (register_class[self.min_label] == 'optimizable'
                     or register_class[self.maj_label] == 'optimizable')):

            target_class = self.determine_target_class(TR=TR,
                                                        GS=GS,
                                                        for_validation=for_validation,
                                                        register_class=register_class)

            # conditional in line 17
            if (target_class == self.min_label
                    and number_of_optimizations[target_class] > 0):
                # this is a tricky part, because GS_trial is defined later only
                GS_trial = self.DE_optimization(GS=GS_trial, # pylint: disable=invalid-name
                                                TR=TR,
                                                indices=indices,
                                                for_validation=for_validation)
            else:
                GS_trial, for_validation_trial = self.generate_trial( # pylint: disable=invalid-name
                                            indices=indices,
                                            target_class=target_class,
                                            TR=TR, GS=GS,
                                            for_validation=for_validation)


                # doing optimization
                GS_trial = self.DE_optimization(GS=GS, # pylint: disable=invalid-name
                                           TR=TR,
                                           indices=indices,
                                           for_validation=for_validation_trial)

            # line 23
            AUC_trial = self.evaluate_ID(GS=GS_trial, # pylint: disable=invalid-name
                                            TR=TrainingSet(TR.X[for_validation_trial],
                                                            TR.y[for_validation_trial]))
            # conditional in line 24
            if AUC_trial > AUC:
                AUC = AUC_trial
                GS = GS_trial
                for_validation = for_validation_trial

                _logger.info("%s: Size of validation set %d",
                        self.__class__.__name__, len(for_validation))
                self.configuration_checks(TR.y, for_validation, GS.y)

                number_of_optimizations[target_class] = 0
            else:
                # conditional in line 29
                if (target_class == self.min_label
                        and number_of_optimizations[target_class] < self.de_params['OT']):
                    number_of_optimizations[target_class] += 1
                else:
                    register_class[target_class] = 'non-optimizable'

        return GS

    def sampling_algorithm(self, X, y):
        """
        Does the sample generation according to the class parameters.

        Args:
            X (np.ndarray): training set
            y (np.array): target labels

        Returns:
            (np.ndarray, np.array): the extended training set and target labels
        """
        mms = MinMaxScaler()
        X = mms.fit_transform(X)

        min_indices = np.where(y == self.min_label)[0]
        maj_indices = np.where(y == self.maj_label)[0]

        indices = {self.min_label: min_indices,
                    self.maj_label: maj_indices}

        try:
            # Phase 1: Initialization
            AUC, for_validation, GS = self.phase_1(TrainingSet(X, y), indices) # pylint: disable=invalid-name

            # Phase 2: Addition of new instances
            GS = self.phase_2(TR=TrainingSet(X, y), # pylint: disable=invalid-name
                            indices=indices,
                            for_validation=for_validation,
                            GS=GS,
                            AUC=AUC)
        except ValueError as value:
            return self.return_copies(X, y, value.args[0])

        return mms.inverse_transform(GS.X), GS.y

    def get_params(self, deep=False):
        """
        Returns:
            dict: the parameters of the current sampling object
        """
        return {'F': self.de_params['F'],
                'G': self.de_params['G'],
                'OT': self.de_params['OT'],
                'max_it': self.de_params['max_it'],
                'n_jobs': self.n_jobs,
                'dt_classifier': self.de_params['dt_classifier'],
                'base_classifier': self.de_params['base_classifier'],
                **OverSampling.get_params(self)}



"""
This module implements the SSO method.
"""

import numpy as np

from sklearn.cluster import KMeans
from sklearn.linear_model import LinearRegression

import scipy.special as sspecial

from ..base import (NearestNeighborsWithMetricTensor,
                                pairwise_distances_mahalanobis)
from ..base import OverSampling
from .._logger import logger
_logger= logger

__all__= ['SSO']

class SSO(OverSampling):
    """
    References:
        * BibTex::

            @InProceedings{sso,
                            author="Rong, Tongwen
                            and Gong, Huachang
                            and Ng, Wing W. Y.",
                            editor="Wang, Xizhao
                            and Pedrycz, Witold
                            and Chan, Patrick
                            and He, Qiang",
                            title="Stochastic Sensitivity Oversampling
                                    Technique for Imbalanced Data",
                            booktitle="Machine Learning and Cybernetics",
                            year="2014",
                            publisher="Springer Berlin Heidelberg",
                            address="Berlin, Heidelberg",
                            pages="161--171",
                            isbn="978-3-662-45652-1"
                            }

    Notes:
        * In the algorithm step 2d adds a constant to a vector. I have
            changed it to a componentwise adjustment, and also used the
            normalized STSM as I don't see any reason why it would be
            some reasonable, bounded value.
    """

    categories = [OverSampling.cat_extensive,
                  OverSampling.cat_uses_classifier,
                  OverSampling.cat_uses_clustering,
                  OverSampling.cat_density_based,
                  OverSampling.cat_metric_learning]

    def __init__(self,
                 proportion=1.0,
                 n_neighbors=5,
                 *,
                 nn_params={},
                 h=10, # pylint: disable=invalid-name
                 n_iter=5,
                 n_jobs=1,
                 random_state=None,
                 **_kwargs):
        """
        Constructor of the sampling object

        Args:
            proportion (float): proportion of the difference of n_maj and n_min
                                to sample e.g. 1.0 means that after sampling
                                the number of minority samples will be equal to
                                the number of majority samples
            n_neighbors (int): number of neighbors
            nn_params (dict): additional parameters for nearest neighbor calculations, any
                                parameter NearestNeighbors accepts, and additionally use
                                {'metric': 'precomputed', 'metric_learning': '<method>', ...}
                                with <method> in 'ITML', 'LSML' to enable the learning of
                                the metric to be used for neighborhood calculations
            h (int): number of hidden units
            n_iter (int): optimization steps
            n_jobs (int): number of parallel jobs
            random_state (int/RandomState/None): initializer of random_state,
                                                    like in sklearn
        """
        super().__init__(random_state=random_state)
        self.check_greater_or_equal(proportion, "proportion", 0)
        self.check_greater_or_equal(n_neighbors, "n_neighbors", 1)
        self.check_greater_or_equal(h, "h", 1)
        self.check_greater_or_equal(n_iter, "n_iter", 1)
        self.check_n_jobs(n_jobs, 'n_jobs')

        self.proportion = proportion
        self.n_neighbors = n_neighbors
        self.nn_params = nn_params
        self.h = h # pylint: disable=invalid-name
        self.n_iter = n_iter
        self.n_jobs = n_jobs

    @ classmethod
    def parameter_combinations(cls, raw=False):
        """
        Generates reasonable parameter combinations.

        Returns:
            list(dict): a list of meaningful parameter combinations
        """
        parameter_combinations = {'proportion': [0.1, 0.25, 0.5, 0.75,
                                                 1.0, 1.5, 2.0],
                                  'n_neighbors': [3, 5],
                                  'h': [2, 5, 10, 20],
                                  'n_iter': [5]}
        return cls.generate_parameter_combinations(parameter_combinations, raw)

    #def eq_6(self, Q, w, u, v, x, h):
    #    """
    #    Equation 6 in the paper
    #    """
    #    tmp_sum = np.zeros(h)
    #    for i in range(h):
    #        a = (x - u[i] + Q)/np.sqrt(2*v[i])
    #        b = (x - u[i] - Q)/np.sqrt(2*v[i])
    #        tmp_prod = (sspecial.erf(a) - sspecial.erf(b))
    #        tmp_sum[i] = np.sqrt(np.pi/2)*v[i]*np.prod(tmp_prod)
    #    return np.dot(tmp_sum, w)/(2*Q)**len(x)

    def eq_6_vectorized(self, *, Q, w, u, v, X): # pylint: disable=invalid-name
        """
        Vectorized implementation of Equation 6 in the paper

        Args:
            Q (np.array): the Q value
            w (np.array): the w vector
            u (np.array): the u vectors
            v (np.array): the v vector
            X (np.array): all training vectors

        Returns:
            np.array: eq. 6 for each X vector
        """
        tmp_a = (X[:, None] - u + Q) / np.sqrt(2 * v)[None, :, None]
        tmp_b = (X[:, None] - u - Q) / np.sqrt(2 * v)[None, :, None]
        tmp_prod = sspecial.erf(tmp_a) - sspecial.erf(tmp_b)
        tmp_sum = (np.sqrt(np.pi / 2.0) * v * np.prod(tmp_prod, axis=2))
        result = np.dot(tmp_sum, w) / (2 * Q) ** X.shape[1]
        return result

    #def eq_8(self, Q, w, u, v, x, h):
    #    """
    #    Equation 8 in the paper
    #    """
    #    res = 0.0
    #    for i in range(h):
    #        vi2 = v[i]**2
    #        for r in range(h):
    #            vr2 = v[r]**2
    #            a1 = (np.sqrt(2*vi2*vr2*(vi2 + vr2)))
    #
    #            a00_v = (vi2 + vr2)*(x + Q)
    #            a01_v = vi2*u[r] + vr2*u[i]
    #            a0_v = a00_v - a01_v
    #            a_v = a0_v/a1
    #
    #            b_v = ((vi2 + vr2)*(x - Q) - (vi2*u[r] + vr2*u[i]))/a1
    #            tmp_prod = sspecial.erf(a_v) - sspecial.erf(b_v)
    #
    #            tmp_a = (np.sqrt(2*vi2*vr2*(vi2 + vr2)) /
    #                        (vi2 + vr2))**len(x)
    #            norm = np.linalg.norm(u[r] - u[i])
    #            tmp_b = np.exp(-0.5 * norm**2/(vi2 + vr2))
    #
    #            res = res + tmp_a*tmp_b*np.prod(tmp_prod)*w[i]*w[r]
    #
    #    return (np.sqrt(np.pi)/(4*Q))**len(x)*res

    def eq_8_av_bv(self, *, Q, u, X, v2, vi_vr, a1):
        """
        Calculate the a_v and b_v arrays.

        Args:
            Q (np.array): the Q value
            u (np.array): the u vectors
            X (np.array): all training vectors
            v2 (np.array): the squared v values
            vi_vr (np.array): the cross sum of squared v values
            a1 (np.array): the precalculated a1 term

        Returns:
            np.array: eq. 8 for each X vector
        """

        a00_v = vi_vr[:, :, None, None] * (X + Q)

        tmp1 = np.repeat(v2, u.shape[0]).reshape(v2.shape[0], u.shape[0])[:, :, None]
        tmp2 = np.array([u] * u.shape[0])

        a01_v = u[:, None] * v2[:, None] + tmp2 * tmp1
        a0_v = a00_v - a01_v[:, :, None]
        a_v = a0_v / a1[:, :, None, None]

        b_v = vi_vr[:, :, None, None] * (X - Q)
        b_v = b_v - a01_v[:, :, None]
        b_v = b_v / a1[:, :, None, None]

        return a_v, b_v

    def eq_8_vectorized(self, *, Q, w, u, v, X): # pylint: disable=invalid-name
        """
        Vectorized implementation of Equation 8 in the paper

        Args:
            Q (np.array): the Q value
            w (np.array): the w vector
            u (np.array): the u vectors
            v (np.array): the v vector
            X (np.array): all training vectors

        Returns:
            np.array: eq. 8 for each X vector
        """
        v2 = v ** 2.0 # pylint: disable=invalid-name

        vi_vr = (v2[:, None] + v2)

        a1 = np.sqrt(2.0 * (v2[:, None] * v2) * vi_vr) # pylint: disable=invalid-name

        a_v, b_v = self.eq_8_av_bv(Q=Q, u=u, X=X, v2=v2, vi_vr=vi_vr, a1=a1)

        tmp_prod = sspecial.erf(a_v) - sspecial.erf(b_v)

        tmp_a = (a1 / vi_vr) ** X.shape[1]

        tmp_b = np.exp(-0.5 * np.linalg.norm(u[:, None] - u, axis=2)**2/vi_vr)

        result = (tmp_a * tmp_b * (w[:, None] * w))[:, :, None] \
                                        * np.prod(tmp_prod, axis=3)

        result = np.sum(np.sum(result, axis=0), axis=0)

        result = (np.sqrt(np.pi) / (4 * Q)) ** X.shape[1] * result

        return result

    def calculate_vectors(self, X, y, X_min, nn_params):
        """
        Calculate the vectors used in the rest of the algorithm.

        Args:
            X (np.array): all training vectors
            y (np.array): all target labels
            X_min (np.array): all minority samples
            nn_params (dict): the nearest neighbor parameters

        Returns:
            np.array, np.array, np.array, np.array: the u, v, f and w
                                                    vectors
        """
        # applying kmeans clustering to find the hidden neurons
        h = min([self.h, len(X_min)]) # pylint: disable=invalid-name
        kmeans = KMeans(n_clusters=h,
                        random_state=self._random_state_init)
        kmeans.fit(X)

        # extracting the hidden center elements
        u = kmeans.cluster_centers_ # pylint: disable=invalid-name

        # extracting scale parameters as the distances of closest centers
        nn_cent = NearestNeighborsWithMetricTensor(n_neighbors=2,
                                                    n_jobs=self.n_jobs,
                                                    **nn_params)
        nn_cent.fit(u)
        dist_cent, _ = nn_cent.kneighbors(u)
        v = dist_cent[:, 1] # pylint: disable=invalid-name

        # computing the response of the hidden units
        phi = pairwise_distances_mahalanobis(X, Y=u, tensor=nn_params.get('metric_tensor', None))
        phi = phi**2
        phi = np.exp(-phi/v**2)

        # applying linear regression to find the best weights
        linreg = LinearRegression()
        linreg.fit(phi, y)
        f = linreg.predict(phi[np.where(y == self.min_label)[0]]) # pylint: disable=invalid-name
        w = linreg.coef_ # pylint: disable=invalid-name

        return u, v, f, w

    def determine_Q(self, X, X_min, nn_params): # pylint: disable=invalid-name
        """
        Calculate the Q value.

        Args:
            X (np.array): all training vectors
            X_min (np.array): all minority samples
            nn_params (dict): the nearest neighbor parameters

        Returns:
            float: the Q value
        """
        # applying nearest neighbors to extract Q values
        n_neighbors = min([self.n_neighbors + 1, len(X)])
        nnmt = NearestNeighborsWithMetricTensor(n_neighbors=n_neighbors,
                                                n_jobs=self.n_jobs,
                                                **nn_params)
        nnmt.fit(X)
        dist, _ = nnmt.kneighbors(X_min)

        Q = np.mean(dist[:, n_neighbors-1])/np.sqrt(len(X[0])) # pylint: disable=invalid-name

        return Q

    def generate_samples(self, X_min, weights, Q, samp_per_iter): # pylint: disable=invalid-name
        """
        Generate samples.

        Args:
            X_min (np.array): the minority samples
            weights (np.array): the density to be used
            Q (float): the Q value
            samp_per_iter (int): the number of samples to generate

        Returns:
            np.array: the generated samples
        """
        base_indices = self.random_state.choice(np.arange(X_min.shape[0]),
                                                samp_per_iter,
                                                p=weights)
        base_vectors = X_min[base_indices]
        lam = (self.random_state.random_sample(base_vectors.shape).T \
                * (2*(1 - weights[base_indices])) - (1 - weights[base_indices])).T

        samples = base_vectors + Q * lam

        return samples

    def determine_Q_and_weights(self, X, y, X_min, nn_params): # pylint: disable=invalid-name
        """
        Determine the Q value and the sampling weights.

        Args:
            X (np.array): all training vectors
            y (np.array): all target labels
            X_min (np.array): all minority samples
            nn_params (dict): the nearest neighbor parameters

        Returns:
            float, np.array: the Q value and the sampling weights
        """
        u, v, f, w = self.calculate_vectors(X, y, X_min, nn_params) # pylint: disable=invalid-name

        Q = self.determine_Q(X, X_min, nn_params) # pylint: disable=invalid-name

        # calculating the sensitivity factors
        I_1 = self.eq_6_vectorized(Q=Q, w=w, u=u, v=v, X=X_min) # pylint: disable=invalid-name
        I_2 = self.eq_8_vectorized(Q=Q, w=w, u=u, v=v, X=X_min) # pylint: disable=invalid-name

        stsm = f**2 - 2*f*I_1 + I_2

        # calculating the sampling weights
        weights = np.abs(stsm)/np.sum(np.abs(stsm))

        return Q, weights

    def sampling_algorithm(self, X, y):
        """
        Does the sample generation according to the class parameters.

        Args:
            X (np.ndarray): training set
            y (np.array): target labels

        Returns:
            (np.ndarray, np.array): the extended training set and target labels
        """
        # number of samples to generate in each iteration
        n_to_sample = self.det_n_to_sample(self.proportion)

        samp_per_iter = max([1, int(n_to_sample/self.n_iter)])

        nn_params = {**self.nn_params}
        nn_params['metric_tensor'] = \
            self.metric_tensor_from_nn_params(nn_params, X, y)

        # executing the algorithm
        for _ in range(self.n_iter):
            X_min = X[y == self.min_label]

            Q, weights = self.determine_Q_and_weights(X, y, X_min, nn_params) # pylint: disable=invalid-name

            samples = self.generate_samples(X_min, weights, Q, samp_per_iter)

            X = np.vstack([X, samples])
            y = np.hstack([y, np.repeat(self.min_label, len(samples))])

        return X.copy(), y.copy()

    def get_params(self, deep=False):
        """
        Returns:
            dict: the parameters of the current sampling object
        """
        return {'proportion': self.proportion,
                'n_neighbors': self.n_neighbors,
                'nn_params': self.nn_params,
                'h': self.h,
                'n_iter': self.n_iter,
                'n_jobs': self.n_jobs,
                **OverSampling.get_params(self)}



"""
This module tests the UnderSampling class.
"""

from smote_variants.base import UnderSampling

def test_undersampling():
    """
    Testing the UnderSampling class.
    """
    unders = UnderSampling()

    assert unders.sample(None, None)[0] is None

    assert len(unders.get_params()) > 0

    assert len(unders.descriptor()) > 0



"""
This module lists some additional objects for testing (with special
parameterizations).
"""

import smote_variants as sv

additional_objs = [sv.Stefanowski(strategy='weak_amp', random_state=5),
                    sv.Stefanowski(strategy='weak_amp_relabel', random_state=5),
                    sv.Stefanowski(strategy='strong_amp', random_state=5),
                    sv.G_SMOTE(method='non-linear_2.0', random_state=5),
                    sv.SMOTE_PSOBAT(method='pso', random_state=5),
                    sv.AHC(strategy='maj', random_state=5),
                    sv.AHC(strategy='minmaj', random_state=5),
                    sv.SOI_CJ(method='jittering', random_state=5),
                    sv.ADG(kernel='rbf_1', random_state=5),
                    sv.SMOTE_IPF(voting='consensus', random_state=5),
                    sv.ASMOBD(smoothing='sigmoid', random_state=5)]



"""
Plotting functionalities.
"""

import warnings

import numpy as np
import pandas as pd
import seaborn as sns
import matplotlib.pyplot as plt

from ..config import suppress_external_warnings

__all__ = ['vector_labeling',
            'plot_raw',
            'plot_oversampled',
            'plot_sampling',
            'plot_sampling_raw',
            'plot_comparison',
            'plot_comparison_raw',
            'SMOTE_VARIANTS_PALETTE',
            'FIG_SIZE_1_1',
            'FIG_SIZE_1_2',
            'FIG_SIZE_1_3',
            'SIZE_SCALE',
            'set_seaborn_theme',
            'relabel_y']

SMOTE_VARIANTS_PALETTE = ['orange', 'olive', 'red']

FIG_SIZE_1_1 = (4.5, 4.5)
FIG_SIZE_1_2 = (10, 4.5)
FIG_SIZE_1_3 = (15, 4.5)

SIZE_SCALE = (100, 300)

def set_seaborn_theme(n_labels):
    """
    Set the seaborn style

    Args:
        n_labels (int): the number of unique class

    Returns:
        dict: the style parameters
    """
    if n_labels <= 3:
        return {'style': 'white',
                'palette': SMOTE_VARIANTS_PALETTE}

    return {'style': 'white'}

def relabel_y(y):
    """
    Relabel the target labels

    Args:
        y (np.array): target labels

    Returns:
        np.array: relabeled labels
    """
    y_new = np.array(['dummy'] * len(y), dtype='object')

    unique_labels = np.unique(y)

    if len(unique_labels) == 2:
        y_new[y == 0] = 'majority'
        y_new[y == 1] = 'minority'
    else:
        y_new[y == 0] = 'majority'
        for idx, label in enumerate(unique_labels):
            if label != 0:
                y_new[y == label] = f'minority {idx}'

    return y_new

def unique_data(X, y, coordinates=(0, 1)):
    """
    Create unique dataset

    Args:
        X (np.array): all training vectors
        y (np.array): all target labels
        coordinates (tuple): the coordinates to use

    Returns:
        pd.DataFrame: the unique data
    """
    coords = [f'coordinate {idx}' for idx in coordinates]

    unified = np.round(np.vstack([X.T, y.T]).T, 4)
    unique, counts = np.unique(unified, return_counts=True, axis=0)
    X_orig, y_orig = unique[:, :-1], unique[:, -1]

    orig_pdf = pd.DataFrame(X_orig[:, coordinates], columns=coords)
    orig_pdf['y'] = y_orig
    orig_pdf['counts'] = counts

    return orig_pdf

def vector_labeling(X, y, X_samp, y_samp, coordinates=(0, 1)):
    """
    Prepares the data for plotting.

    Args:
        X (np.array): the original vectors
        y (np.array): the original target labels
        X_samp (np.array): the new vectors
        y_samp (np.array): the new target labels
        coordinates (tuple): the coordinates to consider

    Returns:
        pd.DataFrame: the data prepared for plotting
    """
    coords = [f'coordinate {idx}' for idx in coordinates]

    orig_pdf = unique_data(X, y, coordinates)
    new_pdf = unique_data(X_samp, y_samp, coordinates)

    orig_pdf = orig_pdf.rename({'counts': 'counts_old'}, axis='columns')

    orig_pdf['indicator_orig'] = 'original'
    new_pdf['indicator_new'] = 'new'

    result = new_pdf.merge(orig_pdf, how='outer', on=(coords + ['y']))

    result.loc[(result['counts_old'] == result['counts']) \
                & (result['indicator_orig'] == 'original') \
                & (result['indicator_new'] == 'new'), 'flag'] = 'untouched'
    result.loc[(result['indicator_orig'].isna()) \
                & (result['indicator_new'] == 'new'), 'flag'] = 'new'
    result.loc[(result['indicator_orig'] == 'original') \
                & (result['indicator_new'].isna()), 'flag'] = 'removed'
    result.loc[(result['counts_old'] != result['counts']) \
                & (result['indicator_orig'] == 'original') \
                & (result['indicator_new'] == 'new'), 'flag'] = 'count_changed'

    result['removed_flag'] = result['flag'] == 'removed'
    result['original_flag'] = result['indicator_orig'] == 'original'
    result['resampled_flag'] = result['flag'].isin(['new', 'untouched'])
    result['new_flag'] = result['flag'] == 'new'
    result['untouched_flag'] = result['flag'] == 'untouched'
    result['count_changed_flag'] = result['flag'] == 'count_changed'

    return result

def plot_raw(X, y, title="", coordinates=(0, 1)):
    """
    Plot a raw distribution

    Args:
        X (np.array): the training vectors
        y (np.array): the target labels
        title (str): the title
        coordinates (tuple): the coordinates to plot

    Returns:
        obj: the figure
    """
    with warnings.catch_warnings():
        if suppress_external_warnings():
            warnings.simplefilter("ignore")

        sns.set_theme(**set_seaborn_theme(len(np.unique(y))))

        fig, axis = plt.subplots(1, 1, figsize=FIG_SIZE_1_1)

        columns = [f'coordinate {idx}' for idx in coordinates]

        data = pd.DataFrame(X[:, coordinates], columns=columns)
        data['y'] = relabel_y(y)

        order = sorted(np.unique(data['y']))

        sns.scatterplot(data=data,
                        x=columns[0],
                        y=columns[1],
                        hue="y",
                        hue_order=order,
                        legend=True,
                        s=SIZE_SCALE[0])
        axis.set_title(title)

        sns.move_legend(axis, "upper left", bbox_to_anchor=(1.0, 1.0))

    return fig

def plot_oversampled(X, y, X_samp, y_samp, *, title="", coordinates=(0, 1)):
    """
    Plot a raw distribution

    Args:
        X (np.array): the training vectors
        y (np.array): the target labels
        X_samp (np.array): the oversampled vectors
        y_samp (np.array): the oversampled labels
        title (str): the title
        coordinates (tuple): the coordinates to plot

    Returns:
        obj: the figure
    """
    _ = y

    with warnings.catch_warnings():
        if suppress_external_warnings():
            warnings.simplefilter("ignore")

        sns.set_theme(**set_seaborn_theme(len(np.unique(y))))

        fig, axis = plt.subplots(1, 1, figsize=FIG_SIZE_1_1)

        columns = [f'coordinate {idx}' for idx in coordinates]

        data = pd.DataFrame(X_samp[:, coordinates], columns=columns)
        data['y'] = y_samp
        data['flag'] = 'original'
        data.loc[X.shape[0]:, 'flag'] = 'new'
        data['y'] = relabel_y(data['y'].values)

        order = np.unique(data['y'].values)

        sns.scatterplot(data=data,
                        x=columns[0],
                        y=columns[1],
                        hue="y",
                        hue_order=order,
                        style='flag',
                        legend=True,
                        s=SIZE_SCALE[0])
        axis.set_title(title)

        sns.move_legend(axis, "upper left", bbox_to_anchor=(1.0, 1.0))

    return fig

def plot_sampling(labeling, title=""):
    """
    Plots the data.

    Args:
        labeling (pd.DataFrame): data prepared for plotting
        title (str): the title

    Returns:
        obj: the figure
    """

    coords = [c for c in labeling.columns if c.startswith('coordinate')]

    with warnings.catch_warnings():
        if suppress_external_warnings():
            warnings.simplefilter("ignore")

        sizes = SIZE_SCALE

        labeling['y'] = relabel_y(labeling['y'].values)

        order = sorted(np.unique(labeling['y'].values))

        sns.set_theme(**set_seaborn_theme(len(order)))

        fig, axes = plt.subplots(1, 2, figsize=FIG_SIZE_1_2)

        axis = sns.scatterplot(data=labeling[labeling['original_flag']],
                                x=coords[0],
                                y=coords[1],
                                hue="y",
                                size="counts_old",
                                hue_order=order,
                                sizes=sizes, legend=False, ax=axes[0])
        axis.set_title("Original dataset")

        axis = sns.scatterplot(data=labeling[labeling['flag'] != 'removed'],
                                x=coords[0],
                                y=coords[1],
                                hue="y",
                                size="counts",
                                hue_order=order,
                                style='flag',
                                sizes=sizes,
                                legend='brief',
                                ax=axes[1],
                                style_order=['untouched', 'new', 'count_changed'])
        sns.scatterplot(data=labeling[labeling['flag'] == 'removed'],
                        x=coords[0],
                        y=coords[1],
                        hue="y",
                        hue_order=order,
                        legend=False,
                        alpha=0.3,
                        ax=axes[1])

        sns.move_legend(axis, "upper left", bbox_to_anchor=(1.0, 1.0))

        plt.title(title)

    return fig

def plot_sampling_raw(*, X, y, X_samp, y_samp, title, coordinates=(0, 1)):
    """
    Plot the sampling from the raw data

    Args:
        X (np.array): original training vectors
        y (np.array): original target labels
        X_samp (np.array): oversampled vectors
        y_samp (np.array): oversampled labels
        title (str): the title of the oversampling
        coordinates (tuple): the coordinates to plot
    """
    labeling = vector_labeling(X, y, X_samp, y_samp, coordinates=coordinates)

    return plot_sampling(labeling, title)

def plot_comparison(labeling0, labeling1, title0, title1):
    """
    Plot the results of two oversamplers against a base one

    Args:
        labeling0 (pd.DataFrame): the vector labeling of the first comparison
        labeling1 (pd.DataFrame): the vector labeling of the second comparison
        title0 (str): title of the first plot
        title1 (str): title of the second plot

    Returns:
        obj: the figure
    """
    coords = [c for c in labeling0.columns if c.startswith('coordinate')]

    labeling0['y'] = relabel_y(labeling0['y'].values)
    labeling1['y'] = relabel_y(labeling1['y'].values)

    with warnings.catch_warnings():
        if suppress_external_warnings():
            warnings.simplefilter("ignore")

        sizes = SIZE_SCALE

        order = sorted(np.unique(labeling0['y']))

        sns.set_theme(**set_seaborn_theme(len(order)))

        fig, axes = plt.subplots(1, 3, figsize=FIG_SIZE_1_3)

        axis = sns.scatterplot(data=labeling0[labeling0['original_flag']],
                                x=coords[0],
                                y=coords[1],
                                hue="y",
                                size="counts_old",
                                hue_order=order,
                                sizes=sizes, legend=False, ax=axes[0])
        axis.set_title("Original dataset")

        axis = sns.scatterplot(data=labeling0[labeling0['flag'] != 'removed'],
                                x=coords[0],
                                y=coords[1],
                                hue="y",
                                size="counts",
                                hue_order=order,
                                style='flag',
                                sizes=sizes,
                                legend=False,
                                ax=axes[1],
                                style_order=['untouched', 'new', 'count_changed'])
        sns.scatterplot(data=labeling0[labeling0['flag'] == 'removed'],
                        x=coords[0],
                        y=coords[1],
                        hue="y",
                        hue_order=order,
                        legend=False,
                        alpha=0.3,
                        ax=axes[1])

        axis.set_title(title0)

        axis = sns.scatterplot(data=labeling1[labeling1['flag'] != 'removed'],
                                x=coords[0],
                                y=coords[1],
                                hue="y",
                                size="counts",
                                hue_order=order,
                                style='flag',
                                sizes=sizes,
                                legend='brief',
                                ax=axes[2],
                                style_order=['untouched', 'new', 'count_changed'])
        sns.scatterplot(data=labeling1[labeling1['flag'] == 'removed'],
                        x=coords[0],
                        y=coords[1],
                        hue="y",
                        hue_order=order,
                        legend=False,
                        alpha=0.3,
                        ax=axes[2])

        sns.move_legend(axis, "upper left", bbox_to_anchor=(1.0, 1.0))
        axis.set_title(title1)

    return fig

def plot_comparison_raw(*, X, y,
                        X_samp0, # pylint: disable=invalid-name
                        y_samp0,
                        X_samp1, # pylint: disable=invalid-name
                        y_samp1,
                        title0,
                        title1,
                        coordinates=(0, 1)):
    """
    Plot the results of two oversamplers against a base one from raw data

    Args:
        X (np.array): the base vectors
        y (np.array): the target labels
        X_samp0 (np.array): the first oversampling
        y_samp0 (np.array): the first target labels
        X_samp1 (np.array): the second oversampling
        y_samp1 (np.array): the second target labels
        title0 (str): the title of the first plot
        title1 (str): the title of the second plot
        coordinates (list): the coordinates to plot

    Returns:
        obj: the figure
    """
    labeling0 = vector_labeling(X, y, X_samp0, y_samp0, coordinates)
    labeling1 = vector_labeling(X, y, X_samp1, y_samp1, coordinates)

    return plot_comparison(labeling0, labeling1, title0, title1)



"""
Testing the DBSMOTE.
"""

from smote_variants import DBSMOTE

from smote_variants.datasets import (load_all_min_noise,
                                     load_3_min_some_maj,
                                     load_separable)

def test_specific():
    """
    Oversampler specific testing
    """

    obj = DBSMOTE(db_iter_limit=1)

    dataset = load_3_min_some_maj()

    X_samp, _ = obj.sample(dataset['data'],
                            dataset['target'])

    assert len(X_samp) > 0

    obj = DBSMOTE(db_iter_limit=1)

    dataset = load_all_min_noise()

    X_samp, _ = obj.sample(dataset['data'],
                            dataset['target'])

    assert len(X_samp) > 0

    obj = DBSMOTE(db_iter_limit=1)

    dataset = load_separable()

    X_samp, _ = obj.sample(dataset['data'],
                            dataset['target'])

    assert len(X_samp) > 0



"""
Test the sampling.
"""

import os
import shutil

import numpy as np

from sklearn import datasets

from smote_variants.evaluation import Folding, SamplingJob
import smote_variants

DATASET = datasets.load_breast_cancer()
DATASET['name'] = 'breast_cancer'

DATASET_WARNING = {'data': np.array([[1, 2], [2, 3],
                                     [1, 2], [2, 3],
                                     [1, 2], [2, 3],
                                     [1, 2], [2, 3],
                                     [1, 2], [2, 3],
                                     [1, 2], [2, 3],
                                     [1, 2], [2, 3],
                                     [1, 2], [2, 3]]),
                   'target': np.array([0, 1,
                                      0, 1,
                                      0, 1,
                                      0, 1,
                                      0, 1,
                                      0, 1,
                                      0, 1,
                                      0, 1]),
                   'name': 'warning dataset'}

cache_path = os.path.join('.', 'test_path')

def test_sampling():
    """
    Testing the sampling.
    """
    shutil.rmtree(cache_path, ignore_errors=True)

    folding = Folding(DATASET,
                    cache_path=cache_path,
                    reset=True)
    oversampler = 'SMOTE'
    oversampler_params = {}

    for fold in folding.fold():
        sjob = SamplingJob(fold,
                        oversampler,
                        oversampler_params,
                        cache_path=cache_path)

        result = sjob.do_oversampling()
        assert os.path.exists(result)

    for fold in folding.fold():
        sjob = SamplingJob(fold,
                        oversampler,
                        oversampler_params,
                        cache_path=cache_path)

        result = sjob.do_oversampling()
        assert os.path.exists(result)

    shutil.rmtree(cache_path)

def test_sampling_warning():
    """
    Testing the sampling with warning.
    """
    smote_variants.config.suppress_internal_warnings(False)

    folding = Folding(DATASET_WARNING,
                    cache_path=cache_path,
                    validator_params={'n_splits': 2, 'n_repeats': 1},
                    reset=True)
    oversampler = 'SMOTE'
    oversampler_params = {}

    for fold in folding.fold():
        sjob = SamplingJob(fold,
                        oversampler,
                        oversampler_params)

        result = sjob.do_oversampling()

        assert len(result['warning']) > 3

    smote_variants.config.suppress_internal_warnings(True)

def test_sampling_error():
    """
    Testing the sampling with error.
    """
    folding = Folding(DATASET,
                    cache_path=cache_path,
                    reset=True)

    for fold in folding.fold():
        oversampler = 'NoSMOTE'
        oversampler_params = {'raise_value_error': True}

        sjob = SamplingJob(fold,
                        oversampler,
                        oversampler_params)

        result = sjob.do_oversampling()

        assert len(result['error']) > 3

        oversampler = 'NoSMOTE'
        oversampler_params = {'raise_runtime_error': True}

        sjob = SamplingJob(fold,
                        oversampler,
                        oversampler_params)

        result = sjob.do_oversampling()

        assert len(result['error']) > 3

def test_sampling_cached_folding():
    """
    Testing the sampling with cached folding.
    """
    shutil.rmtree(cache_path, ignore_errors=True)

    folding = Folding(DATASET,
                    cache_path=cache_path,
                    reset=True)
    oversampler = 'SMOTE'
    oversampler_params = {}

    for fold in folding.folding_files():
        sjob = SamplingJob(fold,
                        oversampler,
                        oversampler_params,
                        cache_path=cache_path)

        result = sjob.do_oversampling()
        assert os.path.exists(result)

    for fold in folding.folding_files():
        sjob = SamplingJob(fold,
                        oversampler,
                        oversampler_params,
                        cache_path=cache_path)

        result = sjob.do_oversampling()
        assert os.path.exists(result)

    shutil.rmtree(cache_path)

def test_sampling_no_caching():
    """
    Testing the sampling without caching.
    """
    folding = Folding(DATASET,
                    reset=True)
    oversampler = 'SMOTE'
    oversampler_params = {}

    for fold in folding.fold():
        sjob = SamplingJob(fold,
                        oversampler,
                        oversampler_params)

        result = sjob.do_oversampling()
        assert isinstance(result, dict)

def test_get_params():
    """
    Testing the get_params function.
    """
    sampling = SamplingJob({},
                    cache_path=None,
                    oversampler = 'SMOTE',
                    oversampler_params = {})
    assert len(sampling.get_params()) > 0



"""
This module implements the ADOMS method.
"""

import numpy as np

from sklearn.decomposition import PCA

from ..base import NearestNeighborsWithMetricTensor, coalesce
from ..base import OverSampling

from .._logger import logger
_logger= logger

__all__= ['ADOMS']

class ADOMS(OverSampling):
    """
    References:
        * BibTex::

            @INPROCEEDINGS{adoms,
                            author={Tang, S. and Chen, S.},
                            booktitle={2008 International Conference on
                                        Information Technology and
                                        Applications in Biomedicine},
                            title={The generation mechanism of synthetic
                                    minority class examples},
                            year={2008},
                            volume={},
                            number={},
                            pages={444-447},
                            keywords={medical image processing;
                                        generation mechanism;synthetic
                                        minority class examples;class
                                        imbalance problem;medical image
                                        analysis;oversampling algorithm;
                                        Principal component analysis;
                                        Biomedical imaging;Medical
                                        diagnostic imaging;Information
                                        technology;Biomedical engineering;
                                        Noise generators;Concrete;Nearest
                                        neighbor searches;Data analysis;
                                        Image analysis},
                            doi={10.1109/ITAB.2008.4570642},
                            ISSN={2168-2194},
                            month={May}}
    """

    categories = [OverSampling.cat_dim_reduction,
                  OverSampling.cat_extensive,
                  OverSampling.cat_metric_learning]

    def __init__(self,
                 proportion=1.0,
                 n_neighbors=5,
                 *,
                 nn_params=None,
                 n_jobs=1,
                 random_state=None,
                 **_kwargs):
        """
        Constructor of the sampling object

        Args:
            proportion (float): proportion of the difference of n_maj and
                                n_min to sample e.g. 1.0 means that after
                                sampling the number of minority samples
                                will be equal to the number of majority
                                samples
            n_neighbors (int): parameter of the nearest neighbor component
            nn_params (dict): additional parameters for nearest neighbor calculations, any
                                parameter NearestNeighbors accepts, and additionally use
                                {'metric': 'precomputed', 'metric_learning': '<method>', ...}
                                with <method> in 'ITML', 'LSML' to enable the learning of
                                the metric to be used for neighborhood calculations
            n_jobs (int): number of parallel jobs
            random_state (int/RandomState/None): initializer of random_state,
                                                    like in sklearn
        """
        super().__init__(random_state=random_state, checks=None)

        self.check_greater_or_equal(proportion, 'proportion', 0.0)
        self.check_greater_or_equal(n_neighbors, 'n_neighbors', 1)
        self.check_n_jobs(n_jobs, 'n_jobs')

        self.proportion = proportion
        self.n_neighbors = n_neighbors
        self.nn_params = coalesce(nn_params, {})
        self.n_jobs = n_jobs

    @classmethod
    def parameter_combinations(cls, raw=False):
        """
        Generates reasonable parameter combinations.

        Returns:
            list(dict): a list of meaningful parameter combinations
        """
        parameter_combinations = {'proportion': [0.1, 0.25, 0.5, 0.75,
                                                 1.0, 1.5, 2.0],
                                  'n_neighbors': [3, 5, 7]}
        return cls.generate_parameter_combinations(parameter_combinations, raw)

    def generate_sample_in_neighborhood(self, sample, neighbors):
        """
        Generate artificial sample in a neighborhood.

        Args:
            sample (np.array): one sample point
            neighbors (np.array): the sample's neighbors

        Returns:
            np.array: the generated sample
        """
        # fitting the PCA
        pca = PCA(n_components=1)
        pca.fit(neighbors)

        # extracting the principal direction
        principal_direction = pca.components_[0]

        # do the sampling according to the description in the paper
        random_index = self.random_state.randint(1, len(neighbors))
        random_neighbor = neighbors[random_index]
        diff = np.linalg.norm(random_neighbor - sample)
        rand = self.random_state.random_sample()
        inner_product = np.dot(random_neighbor - sample,
                                principal_direction)
        sign = 1.0 if inner_product > 0.0 else -1.0

        return sample + sign*rand*diff*principal_direction

    def sampling_algorithm(self, X, y):
        """
        Does the sample generation according to the class parameters.

        Args:
            X (np.ndarray): training set
            y (np.array): target labels

        Returns:
            (np.ndarray, np.array): the extended training set and target labels
        """
        n_to_sample = self.det_n_to_sample(self.proportion)

        if n_to_sample == 0:
            return self.return_copies(X, y, "no need for sampling")

        X_min = X[y == self.min_label]

        # fitting nearest neighbors model
        n_neighbors = np.min([len(X_min), self.n_neighbors+1])

        nn_params= {**self.nn_params}
        nn_params['metric_tensor']= self.metric_tensor_from_nn_params(nn_params, X, y)

        nearestn= NearestNeighborsWithMetricTensor(n_neighbors=n_neighbors,
                                                        n_jobs=self.n_jobs,
                                                        **(nn_params))
        nearestn.fit(X_min)
        indices = nearestn.kneighbors(X_min, return_distance=False)

        samples = []
        for _ in range(n_to_sample):
            index = self.random_state.randint(len(X_min))
            neighbors = X_min[indices[index]]
            sample = X_min[index]

            samples.append(self.generate_sample_in_neighborhood(sample,
                                                                neighbors))

        return (np.vstack([X, np.vstack(samples)]),
                np.hstack([y, np.repeat(self.min_label, len(samples))]))

    def get_params(self, deep=False):
        """
        Returns:
            dict: the parameters of the current sampling object
        """
        return {'proportion': self.proportion,
                'n_neighbors': self.n_neighbors,
                'nn_params': self.nn_params,
                'n_jobs': self.n_jobs,
                **OverSampling.get_params(self)}



"""
Testing the EditedNN
"""

import numpy as np

from smote_variants.noise_removal import EditedNearestNeighbors

def test_specific():
    """
    Oversampler specific testing
    """

    obj = EditedNearestNeighbors()

    X_samp, _ = obj.remove_noise(np.array([[1, 1], [2, 2], [3, 3]]),
                                 np.array([0, 0, 1]))

    assert len(X_samp) == 3



"""
This module implements a wrapper over the MLPClassifier
for easier parameterization.
"""

from sklearn.neural_network import MLPClassifier

class MLPClassifierWrapper:
    """
    Wrapper over MLPClassifier of sklearn to provide easier parameterization
    """

    def __init__(self,
                 *,
                 activation='relu',
                 hidden_layer_fraction=0.1,
                 alpha=0.0001,
                 learning_rate='constant',
                 random_state=None):
        """
        Constructor of the MLPClassifier

        Args:
            activation (str): name of the activation function
            hidden_layer_fraction (float): fraction of the hidden neurons of
                                            the number of input dimensions
            alpha (float): alpha parameter of the MLP classifier
            random_state (int/np.random.RandomState/None): initializer of the
                                                            random state
        """
        self.activation = activation
        self.hidden_layer_fraction = hidden_layer_fraction
        self.alpha = alpha
        self.learning_rate= learning_rate
        self.random_state = random_state
        self.model = None

    def fit(self, X, y):
        """
        Fit the model to the data

        Args:
            X (np.ndarray): features
            y (np.array): target labels

        Returns:
            obj: the MLPClassifierWrapper object
        """
        hidden_layer_size = max([1, int(len(X[0])*self.hidden_layer_fraction)])
        self.model = MLPClassifier(activation=self.activation,
                                   hidden_layer_sizes=(hidden_layer_size,),
                                   alpha=self.alpha,
                                   learning_rate=self.learning_rate,
                                   random_state=self.random_state).fit(X, y)
        return self

    def predict(self, X):
        """
        Predicts the labels of the unseen data

        Args:
            X (np.ndarray): unseen features

        Returns:
            np.array: predicted labels
        """
        return self.model.predict(X)

    def predict_proba(self, X):
        """
        Predicts the class probabilities of the unseen data

        Args:
            X (np.ndarray): unseen features

        Returns:
            np.array: predicted class probabilities
        """
        return self.model.predict_proba(X)

    def get_params(self, deep=False):
        """
        Returns the parameters of the classifier.

        Returns:
            dict: the parameters of the object
        """
        _ = deep
        return {'activation': self.activation,
                'hidden_layer_fraction': self.hidden_layer_fraction,
                'alpha': self.alpha,
                'learning_rate': self.learning_rate,
                'random_state': self.random_state}

    def copy(self):
        """
        Creates a copy of the classifier.

        Returns:
            obj: a copy of the classifier
        """
        return MLPClassifierWrapper(**self.get_params())



"""
This module implements the Stefanowski method.
"""

import numpy as np

from scipy.stats import mode

from ..base import coalesce
from ..base import NearestNeighborsWithMetricTensor
from ..base import OverSampling

from .._logger import logger
_logger= logger

__all__= ['Stefanowski']

class Stefanowski(OverSampling):
    """
    References:
        * BibTex::

            @inproceedings{stefanowski,
                 author = {Stefanowski, Jerzy and Wilk, Szymon},
                 title = {Selective Pre-processing of Imbalanced Data for
                            Improving Classification Performance},
                 booktitle = {Proceedings of the 10th International Conference
                                on Data Warehousing and Knowledge Discovery},
                 series = {DaWaK '08},
                 year = {2008},
                 isbn = {978-3-540-85835-5},
                 location = {Turin, Italy},
                 pages = {283--292},
                 numpages = {10},
                 url = {http://dx.doi.org/10.1007/978-3-540-85836-2_27},
                 doi = {10.1007/978-3-540-85836-2_27},
                 acmid = {1430591},
                 publisher = {Springer-Verlag},
                 address = {Berlin, Heidelberg},
                }
    """

    categories = [OverSampling.cat_changes_majority,
                  OverSampling.cat_noise_removal,
                  OverSampling.cat_sample_copy,
                  OverSampling.cat_borderline,
                  OverSampling.cat_metric_learning]

    def __init__(self,
                 *,
                 strategy='weak_amp',
                 nn_params=None,
                 n_jobs=1,
                 random_state=None,
                 **_kwargs):
        """
        Constructor of the sampling object

        Args:
            strategy (str): 'weak_amp'/'weak_amp_relabel'/'strong_amp'
            nn_params (dict): additional parameters for nearest neighbor calculations, any
                                parameter NearestNeighbors accepts, and additionally use
                                {'metric': 'precomputed', 'metric_learning': '<method>', ...}
                                with <method> in 'ITML', 'LSML' to enable the learning of
                                the metric to be used for neighborhood calculations
            n_jobs (int): number of parallel jobs
            random_state (int/RandomState/None): initializer of random_state,
                                                    like in sklearn
        """
        super().__init__(random_state=random_state, checks={'min_n_min': 6})

        self.check_isin(strategy,
                        'strategy',
                        ['weak_amp', 'weak_amp_relabel', 'strong_amp'])
        self.check_n_jobs(n_jobs, 'n_jobs')

        self.strategy = strategy
        self.nn_params = coalesce(nn_params, {})
        self.n_jobs = n_jobs

    @ classmethod
    def parameter_combinations(cls, raw=False):
        """
        Generates reasonable parameter combinations.

        Returns:
            list(dict): a list of meaningful parameter combinations
        """
        if not raw:
            return [{'strategy': 'weak_amp'},
                    {'strategy': 'weak_amp_relabel'},
                    {'strategy': 'strong_amp'}, ]

        return {'strategy': ['weak_amp', 'weak_amp_relabel', 'strong_amp']}

    def create_multiples(self, X, indices, numbers):
        """
        Create multiples from vectors.

        Args:
            X (np.array): base vectors
            indices (np.array): the indices of vectors to multiply
            numbers (np.array): the counts of multiplication

        Returns:
            np.array: the multipled vectors
        """
        samples = np.zeros(shape=(0, X.shape[1]))
        for idx, min_idx in enumerate(indices):
            samples = np.vstack([samples, np.tile(X[min_idx], (numbers[idx], 1))])
        return samples

    def weak_amp(self, *, X, y, safe_flag, indices):
        """
        Weak amplification.

        Args:
            X (np.array): all training vectors
            y (np.array): all target labels
            safe_flag (np.array): True if the vector is safe, False if noise
            indices (np.array): the neighborhood structure

        Returns:
            np.array: the generated samples
        """
        minority_indices = np.where(y == self.min_label)[0]

        samples = np.zeros((0, X.shape[1]))
        # weak mplification - the number of copies is the number of
        # majority nearest neighbors
        mask = ~ safe_flag[minority_indices]
        minority_masked = minority_indices[mask]
        k = np.sum((y[indices[minority_masked][:, 1:]] == self.maj_label) # pylint: disable=invalid-name
                        & (safe_flag[indices[minority_masked][:, 1:]]), axis=1)

        samples = np.vstack([samples,
                             self.create_multiples(X, minority_masked, k)])

        return samples

    def weak_amp_relabel(self, *, y, safe_flag, D, indices): # pylint: disable=invalid-name
        """
        Weak amplification relabeling.

        Args:
            y (np.array): all target labels
            safe_flag (np.array): True if the vector is safe, False if noise
            D (np.array): the D array
            indices (np.array): the neighborhood structure

        Returns:
            np.array, np.array: the relabeled y and the updated D array
        """
        minority_indices = np.where(y == self.min_label)[0]

        # relabling - noisy majority neighbors are relabelled to minority
        mask = ~ safe_flag[minority_indices]
        minority_masked = minority_indices[mask]
        ind_tmp = indices[minority_masked][:, 1:]
        ind_tmp = ind_tmp[(y[ind_tmp] == self.maj_label) & (~ safe_flag[ind_tmp])]
        ind = list(set(ind_tmp.flatten().tolist()))
        y[ind] = self.min_label
        D[ind] = False

        return y, D

    def strong_amp(self, *, X, y, safe_flag, indices, indices5):
        """
        Strong amplification.

        Args:
            X (np.array): all training vectors
            y (np.array): all target labels
            safe_flag (np.array): True if the vector is safe, False if noise
            indices (np.array): the neighborhood structure
            indices5 (np.array): the neighborhood structure with 5 neighbors

        Returns:
            np.array: the generated samples
        """
        minority_indices = np.where(y == self.min_label)[0]

        samples = np.zeros((0, X.shape[1]))
        # safe minority samples are copied as many times as many safe
        # majority samples are among the nearest neighbors
        mask = safe_flag[minority_indices]
        minority_masked = minority_indices[mask]
        k = np.sum((y[indices[minority_masked][:, 1:]] == self.maj_label) # pylint: disable=invalid-name
                        & (safe_flag[indices[minority_masked][:, 1:]]), axis=1)
        samples = np.vstack([samples,
                             self.create_multiples(X, minority_masked, k)])

        mask = ~ safe_flag[minority_indices]
        minority_masked = minority_indices[mask]
        correct_mask = mode(y[indices5[minority_masked, 1:]]).mode[:, 0] \
                                                        == y[minority_masked]

        k = np.sum((y[indices[minority_masked[correct_mask], 1:]] == self.maj_label) \
                    & safe_flag[indices[minority_masked[correct_mask], 1:]], axis=1) # pylint: disable=invalid-name
        samples = np.vstack([samples,
                             self.create_multiples(X, minority_masked[correct_mask], k)])

        k = np.sum((y[indices5[minority_masked[~correct_mask], 1:]] == self.maj_label) \
                    & safe_flag[indices5[minority_masked[~correct_mask], 1:]], axis=1) # pylint: disable=invalid-name
        samples = np.vstack([samples,
                             self.create_multiples(X, minority_masked[~correct_mask], k)])

        return samples

    def sampling_algorithm(self, X, y):
        """
        Does the sample generation according to the class parameters.

        Args:
            X (np.ndarray): training set
            y (np.array): target labels

        Returns:
            (np.ndarray, np.array): the extended training set and target labels
        """
        nn_params = {**self.nn_params}
        nn_params['metric_tensor'] = \
                self.metric_tensor_from_nn_params(nn_params, X, y)

        X_orig, y_orig = X, y

        # copying y as its values will change
        y = y.copy()
        # fitting the nearest neighbors model for noise filtering, 4 neighbors
        # instead of 3 as the closest neighbor to a point is itself
        nnmt= NearestNeighborsWithMetricTensor(n_neighbors=np.min([4, len(X)]),
                                               n_jobs=self.n_jobs,
                                               **nn_params)
        nnmt.fit(X)
        indices = nnmt.kneighbors(X, return_distance=False)

        # fitting the nearest neighbors model for sample generation,
        # 6 neighbors instead of 5 for the same reason
        nnmt= NearestNeighborsWithMetricTensor(n_neighbors=np.min([6, len(X)]),
                                                n_jobs=self.n_jobs,
                                                **nn_params)
        nnmt.fit(X)
        indices5 = nnmt.kneighbors(X, return_distance=False)

        # determining noisy and safe flags
        safe_flag = mode(y[indices[:, 1:]], axis=1).mode[:, 0] == y
        D = (y == self.maj_label) & (~ safe_flag) # pylint: disable=invalid-name

        samples = np.zeros((0, X.shape[1]))

        if self.strategy in ['weak_amp', 'weak_amp_relabel']:
            samples = np.vstack([samples,
                                 self.weak_amp(X=X,
                                                y=y,
                                                safe_flag=safe_flag,
                                                indices=indices)])

        if self.strategy == 'weak_amp_relabel':
            y, D = self.weak_amp_relabel(y=y, # pylint: disable=invalid-name
                                        safe_flag=safe_flag,
                                        D=D,
                                        indices=indices)

        if self.strategy == 'strong_amp':
            samples = np.vstack([samples,
                                 self.strong_amp(X=X,
                                                y=y,
                                                safe_flag=safe_flag,
                                                indices=indices,
                                                indices5=indices5)])

        to_remove = np.where(D)[0]

        X_noise_removed = np.delete(X, to_remove, axis=0) # pylint: disable=invalid-name
        y_noise_removed = np.delete(y, to_remove, axis=0)

        if np.unique(y_noise_removed).shape[0] < 2:
            return self.return_copies(X_orig, y_orig, "one class has been completely removed")

        return (np.vstack([X_noise_removed, samples]),
                np.hstack([y_noise_removed,
                           np.repeat(self.min_label, len(samples))]))

    def get_params(self, deep=False):
        """
        Returns:
            dict: the parameters of the current sampling object
        """
        return {'strategy': self.strategy,
                'nn_params': self.nn_params,
                'n_jobs': self.n_jobs,
                **OverSampling.get_params(self)}



"""
This module implements the CE_SMOTE method.
"""
import warnings

import numpy as np

from sklearn.cluster import KMeans
import scipy.optimize as soptimize

from ..base import coalesce_dict, coalesce
from ..base import NearestNeighborsWithMetricTensor
from ..base import OverSamplingSimplex
from .._logger import logger
_logger= logger

__all__= ['CE_SMOTE']

class CE_SMOTE(OverSamplingSimplex):
    """
    References:
        * BibTex::

            @INPROCEEDINGS{ce_smote,
                                author={Chen, S. and Guo, G. and Chen, L.},
                                booktitle={2010 IEEE 24th International
                                            Conference on Advanced Information
                                            Networking and Applications
                                            Workshops},
                                title={A New Over-Sampling Method Based on
                                        Cluster Ensembles},
                                year={2010},
                                volume={},
                                number={},
                                pages={599-604},
                                keywords={data mining;Internet;pattern
                                            classification;pattern clustering;
                                            over sampling method;cluster
                                            ensembles;classification method;
                                            imbalanced data handling;CE-SMOTE;
                                            clustering consistency index;
                                            cluster boundary minority samples;
                                            imbalanced public data set;
                                            Mathematics;Computer science;
                                            Electronic mail;Accuracy;Nearest
                                            neighbor searches;Application
                                            software;Data mining;Conferences;
                                            Web sites;Information retrieval;
                                            classification;imbalanced data
                                            sets;cluster ensembles;
                                            over-sampling},
                                doi={10.1109/WAINA.2010.40},
                                ISSN={},
                                month={April}}
    """

    categories = [OverSamplingSimplex.cat_extensive,
                  OverSamplingSimplex.cat_borderline,
                  OverSamplingSimplex.cat_uses_clustering,
                  OverSamplingSimplex.cat_sample_ordinary,
                  OverSamplingSimplex.cat_metric_learning]

    def __init__(self,
                 proportion=1.0,
                 *,
                 h=10,
                 k=5,
                 nn_params=None,
                 ss_params=None,
                 alpha=0.5,
                 n_jobs=1,
                 random_state=None,
                 **_kwargs):
        """
        Constructor of the sampling object

        Args:
            proportion (float): proportion of the difference of n_maj and n_min
                                to sample e.g. 1.0 means that after sampling
                                the number of minority samples will be equal to
                                the number of majority samples
            h (int): size of ensemble
            k (int): number of clusters/neighbors
            nn_params (dict): additional parameters for nearest neighbor calculations, any
                                parameter NearestNeighbors accepts, and additionally use
                                {'metric': 'precomputed', 'metric_learning': '<method>', ...}
                                with <method> in 'ITML', 'LSML' to enable the learning of
                                the metric to be used for neighborhood calculations
            alpha (float): [0,1] threshold to select boundary samples
            n_jobs (int): number of parallel jobs
            random_state (int/RandomState/None): initializer of random_state,
                                                    like in sklearn
        """
        ss_params_default = {'n_dim': 2, 'simplex_sampling': 'uniform',
                            'within_simplex_sampling': 'random',
                            'gaussian_component': None}
        ss_params = coalesce_dict(ss_params, ss_params_default)

        super().__init__(**ss_params, random_state=random_state)
        self.check_greater_or_equal(proportion, "proportion", 0)
        self.check_greater_or_equal(h, "h", 1)
        self.check_greater_or_equal(k, "k", 1)
        self.check_in_range(alpha, "alpha", [0, 1])
        self.check_n_jobs(n_jobs, 'n_jobs')

        self.proportion = proportion
        self.h = h # pylint: disable=invalid-name
        self.k = k # pylint: disable=invalid-name
        self.nn_params = coalesce(nn_params, {})
        self.alpha = alpha
        self.n_jobs = n_jobs

    @ classmethod
    def parameter_combinations(cls, raw=False):
        """
        Generates reasonable parameter combinations.

        Returns:
            list(dict): a list of meaningful parameter combinations
        """
        parameter_combinations = {'proportion': [0.1, 0.25, 0.5, 0.75,
                                                 1.0, 1.5, 2.0],
                                  'h': [5, 10, 15],
                                  'k': [3, 5, 7],
                                  'alpha': [0.2, 0.5, 0.8]}
        return cls.generate_parameter_combinations(parameter_combinations, raw)

    def do_the_clustering(self, X):
        """
        Do the clustering.

        Args:
            X (np.array): all features

        Returns:
            np.array: the labels
        """
        n_dim = X.shape[1]
        labels = []
        for _ in range(self.h):
            n_features = self.random_state.randint(np.max([int(n_dim/2), 1]), n_dim + 1)
            features = self.random_state.choice(n_dim, n_features, replace=False)
            n_clusters = min([len(X), self.k])
            kmeans = KMeans(n_clusters=n_clusters,
                            random_state=self._random_state_init)
            with warnings.catch_warnings():
                warnings.simplefilter('ignore')
                kmeans.fit(X[:, features])
            labels.append(kmeans.labels_)

        return labels

    def cluster_matching(self, labels):
        """
        Do the cluster matching.

        Args:
            labels (list): labelings

        Returns:
            np.array: the cluster matched labeling
        """
        base_label = 0
        for idx, labeling in enumerate(labels):
            if not idx == base_label:
                cost_matrix = np.zeros(shape=(self.k, self.k))

                for jdx in range(self.k):
                    mask_j = labels[base_label] == jdx
                    for kdx in range(self.k):
                        mask_k = labeling == kdx
                        mask_jk = np.logical_and(mask_j, mask_k)
                        cost_matrix[jdx, kdx] = np.sum(mask_jk)

                # solving the assignment problem
                row_ind, _ = soptimize.linear_sum_assignment(-cost_matrix)

                # doing the relabeling
                relabeling = labeling.copy()

                for jdx in range(len(row_ind)):
                    relabeling[labeling == kdx] = jdx

                labels[idx] = relabeling

        return np.vstack(labels)

    def cluster_consistency_index(self, labels):
        """
        Computing the cluster consistency index

        Args:
            np.array: labeling

        Returns:
            np.array: the cluster consistency indices
        """
        cci = np.apply_along_axis(lambda x: max(set(x.tolist()),
                                    key=x.tolist().count), 0, labels)
        cci = np.sum(labels == cci, axis=0)
        cci = cci / self.h

        return cci

    def sampling_algorithm(self, X, y):
        """
        Does the sample generation according to the class parameters.

        Args:
            X (np.ndarray): training set
            y (np.array): target labels

        Returns:
            (np.ndarray, np.array): the extended training set and target labels
        """
        n_to_sample = self.det_n_to_sample(self.proportion)

        if n_to_sample == 0:
            self.return_copies(X, y, "Sampling is not needed")

        # do the clustering and labelling
        labels = self.do_the_clustering(X)

        # do the cluster matching, clustering 0 will be considered the one to
        # match the others to the problem of finding cluster matching is
        # basically the "assignment problem"
        labels = self.cluster_matching(labels)

        # compute clustering consistency index
        cci = self.cluster_consistency_index(labels)

        # determining minority boundary samples
        P_boundary = X[(y == self.min_label) & (cci <= self.alpha)] # pylint: disable=invalid-name

        # there might be no boundary samples
        if len(P_boundary) <= self.n_dim - 1:
            return self.return_copies(X, y, "empty boundary")

        # finding nearest neighbors of boundary samples
        n_neighbors = min([len(P_boundary), self.k])

        nn_params= {**self.nn_params}
        nn_params['metric_tensor']= \
                        self.metric_tensor_from_nn_params(nn_params, X, y)

        nnmt = NearestNeighborsWithMetricTensor(n_neighbors=n_neighbors,
                                                n_jobs=self.n_jobs,
                                                **(nn_params))
        nnmt.fit(P_boundary)
        ind = nnmt.kneighbors(P_boundary, return_distance=False)

        samples = self.sample_simplex(X=P_boundary,
                                        indices=ind,
                                        n_to_sample=n_to_sample)

        # do the sampling
        #samples = []
        #for _ in range(n_to_sample):
        #    idx = self.random_state.randint(len(ind))
        #    point_a = P_boundary[idx]
        #    point_b = P_boundary[self.random_state.choice(ind[idx][1:])]
        #    samples.append(self.sample_between_points(point_a, point_b))

        return (np.vstack([X, samples]),
                np.hstack([y, np.repeat(self.min_label, len(samples))]))

    def get_params(self, deep=False):
        """
        Returns:
            dict: the parameters of the current sampling object
        """
        return {'proportion': self.proportion,
                'h': self.h,
                'k': self.k,
                'nn_params': self.nn_params,
                'alpha': self.alpha,
                'n_jobs': self.n_jobs,
                **OverSamplingSimplex.get_params(self)}
